# encoding=utf8
# The certificate manager

"""The certificate manager
"""

import logging

import os
import os.path
import shutil
import subprocess

import jinja2

class CertificateManager(object):
    """The certificate manager
    """
    DIR_CERTS = 'certs'
    DIR_CRL = 'crl'
    DIR_CSR = 'csr'
    DIR_NEWCERTS = 'newcerts'
    DIR_PRIVATE = 'private'

    DIRS = [ DIR_CERTS, DIR_CRL, DIR_CSR, DIR_NEWCERTS, DIR_PRIVATE ]

    FILE_INDEX = 'index.txt'
    FILE_SERIAL = 'serial'
    FILE_CONFIG = 'openssl.config'

    FILE_CA_CERT = 'ca.cert.pem'
    FILE_CA_KEY = 'ca.key.pem'

    def __init__(self, basePath = None, logger = None):
        """Create a new CertificateManager
        """
        self.basePath = os.path.expanduser(os.path.abspath(basePath)) or os.getcwd()
        self.logger = logger or logging.getLogger('certmgr')
        # Check base path
        if not os.path.isdir(self.basePath):
            raise ValueError('Base path [%s] not found' % self.basePath)

    def init(self):
        """Init the manager
        NOTE:
            This method will remove all files and directories in the base path
        Returns:
            Nothing
        """
        self.logger.info('Start initialize dir [%s]', self.basePath)
        # Clean the directory
        for name in os.listdir(self.basePath):
            path = os.path.join(self.basePath, name)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.unlink(path)
        # Prepare dirs and files
        self.logger.info('Prepare dirs and files')
        for name in self.DIRS:
            os.mkdir(os.path.join(self.basePath, name))
        with open(os.path.join(self.basePath, self.FILE_INDEX), 'wb') as fd:
            pass
        with open(os.path.join(self.basePath, self.FILE_SERIAL), 'wb') as fd:
            print >>fd, 1000
        # Format the config file
        env = jinja2.Environment(
                loader = jinja2.FileSystemLoader(os.path.dirname(os.path.abspath(__file__))),
                trim_blocks = True,
                lstrip_blocks = True
                )
        template = env.get_template('openssl.config.template')
        content = template.render(basePath = self.basePath)
        with open(os.path.join(self.basePath, self.FILE_CONFIG), 'wb') as fd:
            print >>fd, content
        # Done

    def verifyCertificate(self, name):
        """Verify a certificate
        """
        certPath = os.path.join(self.basePath, self.DIR_CERTS, '%s.cert.pem' % name)
        if not os.path.isfile(certPath):
            raise ValueError('Certificate [%s] not found' % certPath)
        if subprocess.call([ 'openssl', 'x509', '-noout', '-text', '-in', certPath ]) != 0:
            raise ValueError('Failed to verify the certificate')

    def createRootCertificate(self, noPass = False, keyLength = 4096):
        """Create a root certificate
        """
        configPath, keyPath, certPath = \
                os.path.join(self.basePath, self.FILE_CONFIG), \
                os.path.join(self.basePath, self.DIR_PRIVATE, self.FILE_CA_KEY), \
                os.path.join(self.basePath, self.DIR_CERTS, self.FILE_CA_CERT)
        if os.path.isfile(keyPath):
            raise ValueError('CA key file exists at [%s]' % keyPath)
        if os.path.isfile(certPath):
            raise ValueError('CA certificate file exists at [%s]' % certPath)
        # Generate the key
        if subprocess.call([ 'openssl', 'genrsa', '-aes256', '-out', keyPath, str(keyLength) ] if not noPass else [ 'openssl', 'genrsa', '-out', keyPath, str(keyLength) ]) != 0:
            raise ValueError('Failed to create root key')
        # Generate the cert
        if subprocess.call([ 'openssl', 'req', '-config', configPath, '-key', keyPath, '-new', '-x509', '-days', '7300', '-sha256', '-extensions', 'v3_ca', '-out', certPath ]) != 0:
            raise ValueError('Failed to create root certifcate')
        # Done

    def createServerCertificate(self, name, noPass = True, keyLength = 2048, days = 375):
        """Create a server certificate
        """
        if not name:
            raise ValueError('Require name')
        configPath, keyPath, csrPath, certPath = \
                os.path.join(self.basePath, self.FILE_CONFIG), \
                os.path.join(self.basePath, self.DIR_PRIVATE, '%s.key.pem' % name), \
                os.path.join(self.basePath, self.DIR_CSR, '%s.csr.pem' % name), \
                os.path.join(self.basePath, self.DIR_CERTS, '%s.cert.pem' % name)
        if os.path.isfile(keyPath):
            raise ValueError('Key file exists at [%s]' % keyPath)
        if os.path.isfile(csrPath):
            raise ValueError('CSR file exists at [%s]' % csrPath)
        if os.path.isfile(certPath):
            raise ValueError('Certificate file exists at [%s]' % certPath)
        # Generate the key
        if subprocess.call([ 'openssl', 'genrsa', '-aes256', '-out', keyPath, str(keyLength) ] if not noPass else [ 'openssl', 'genrsa', '-out', keyPath, str(keyLength) ]) != 0:
            raise ValueError('Failed to create key')
        # Generate the csr
        if subprocess.call([ 'openssl', 'req', '-config', configPath, '-key', keyPath, '-new', '-sha256', '-out', csrPath ]) != 0:
            raise ValueError('Failed to create CSR')
        # Generate the certificate
        if subprocess.call([ 'openssl', 'ca', '-config', configPath, '-extensions', 'server_cert', '-days', str(days), '-notext', '-md', 'sha256', '-in', csrPath, '-out', certPath ]) != 0:
            raise ValueError('Failed to create certificate')
        # Done

    def createClientCertificate(self, name, noPass = True, keyLength = 2048, days = 375):
        """Create a client certificate
        """
        if not name:
            raise ValueError('Require name')
        configPath, keyPath, csrPath, certPath = \
                os.path.join(self.basePath, self.FILE_CONFIG), \
                os.path.join(self.basePath, self.DIR_PRIVATE, '%s.key.pem' % name), \
                os.path.join(self.basePath, self.DIR_CSR, '%s.csr.pem' % name), \
                os.path.join(self.basePath, self.DIR_CERTS, '%s.cert.pem' % name)
        if os.path.isfile(keyPath):
            raise ValueError('Key file exists at [%s]' % keyPath)
        if os.path.isfile(csrPath):
            raise ValueError('CSR file exists at [%s]' % csrPath)
        if os.path.isfile(certPath):
            raise ValueError('Certificate file exists at [%s]' % certPath)
        # Generate the key
        if subprocess.call([ 'openssl', 'genrsa', '-aes256', '-out', keyPath, str(keyLength) ] if not noPass else [ 'openssl', 'genrsa', '-out', keyPath, str(keyLength) ]) != 0:
            raise ValueError('Failed to create key')
        # Generate the csr
        if subprocess.call([ 'openssl', 'req', '-config', configPath, '-key', keyPath, '-new', '-sha256', '-out', csrPath ]) != 0:
            raise ValueError('Failed to create CSR')
        # Generate the certificate
        if subprocess.call([ 'openssl', 'ca', '-config', configPath, '-extensions', 'usr_cert', '-days', str(days), '-notext', '-md', 'sha256', '-in', csrPath, '-out', certPath ]) != 0:
            raise ValueError('Failed to create certificate')
        # Done




# encoding=utf8
# The setup file

import sys
import os.path

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

import calib

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

requirements = [ x.strip() for x in open('requirements.txt').readlines() ]

setup(
    name = 'calib',
    version = calib.__version__,
    author = 'lipixun',
    author_email = 'lipixun@outlook.com',
    url = 'https://github.com/lipixun/pyca',
    packages = [ 'calib' ],
    package_data = { 'calib': [ 'openssl.config.template' ] },
    install_requires = requirements,
    license = 'LICENSE',
    description = 'Certificate Authority Lib & Tools',
    long_description = open('README.md').read(),
    keywords = [ 'python', 'ca', 'certificate', 'openssl' ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    ]
)




# encoding=utf8
# The ca lib & tools

"""The certificate authority lib & tool

This lib is based on openssl and follows the guide: https://jamielinux.com/docs/openssl-certificate-authority/

"""

from version import __version__




# encoding=utf8
# The ca tool entry

"""The ca tool entry
"""

import sys
reload(sys)
sys.setdefaultencoding('utf8')

import logging
import traceback

from argparse import ArgumentParser

from manager import CertificateManager

logging.basicConfig(format = '%(asctime)s - %(levelname)s - %(name)s - %(message)s', level = logging.INFO)
logger = logging.getLogger('utility')

def getArguments():
    """Get the arguments
    """
    parser = ArgumentParser(description = 'CA utility')
    parser.add_argument('--base', dest = 'basePath', default = '.', help = 'The base working directory')
    subParsers = parser.add_subparsers(dest = 'action')
    # The init parser
    initParser = subParsers.add_parser('init', help = 'Initialize the database')
    # Vertify certificatae
    verifyParser = subParsers.add_parser('verify', help = 'Verify certificate')
    verifyParser.add_argument('--name', dest = 'name', required = True, help = 'The certificate name')
    # Create root certificate parser
    createRootCertParser = subParsers.add_parser('createRootCert', help = 'Create root certificate')
    createRootCertParser.add_argument('--nopass', dest = 'noPass', default = False, action = 'store_true', help = 'Generate root key without a password (Not recommended), false by default.')
    createRootCertParser.add_argument('--keylen', dest = 'keyLength', default = '4096', choices = [ '2048', '4096' ], help = 'The root key length, 4096 by default')
    # Create server certificate parser
    createServerCertParser = subParsers.add_parser('createServerCert', help = 'Create server cert')
    createServerCertParser.add_argument('--name', dest = 'name', required = True, help = 'The certificate name')
    createServerCertParser.add_argument('--usepass', dest = 'usePass', default = False, action = 'store_true', help = 'Generate key with a password, false by default.')
    createServerCertParser.add_argument('--keylen', dest = 'keyLength', default = '2048', choices = [ '2048', '4096' ], help = 'The key length, 2048 by default')
    createServerCertParser.add_argument('--days', dest = 'days', type = int, default = 3650, help = 'The certificate days parameter')
    # Create client certificate parser
    createClientCertParser = subParsers.add_parser('createClientCert', help = 'Create server cert')
    createClientCertParser.add_argument('--name', dest = 'name', required = True, help = 'The certificate name')
    createClientCertParser.add_argument('--usepass', dest = 'usePass', default = False, action = 'store_true', help = 'Generate key with a password, false by default.')
    createClientCertParser.add_argument('--keylen', dest = 'keyLength', default = '2048', choices = [ '2048', '4096' ], help = 'The key length, 2048 by default')
    createClientCertParser.add_argument('--days', dest = 'days', type = int, default = 3650, help = 'The certificate days parameter')
    # Done
    return parser.parse_args()

def main():
    """The main entry
    """
    args = getArguments()
    try:
        if args.action == 'init':
            # Ask for init
            while True:
                print 'Initialize the path [%s] will cause any files or dirs be removed, continue?[y/n]' % args.basePath,
                text = raw_input()
                if text.lower() == 'n':
                    print 'Will not initialize the path, exit'
                    return 1
                elif text.lower() == 'y':
                    # Initialize
                    manager = CertificateManager(args.basePath)
                    manager.init()
                    return 0
                else:
                    print 'Invalid input'
        elif args.action == 'verify':
            # Verify the certificate
            manager = CertificateManager(args.basePath)
            manager.verifyCertificate(args.name)
            return 0
        elif args.action == 'createRootCert':
            # Create the root cert
            manager = CertificateManager(args.basePath)
            manager.createRootCertificate(args.noPass, int(args.keyLength))
            return 0
        elif args.action == 'createServerCert':
            # Create the server cert
            manager = CertificateManager(args.basePath)
            manager.createServerCertificate(args.name, not args.usePass, args.keyLength, args.days)
            return 0
        elif args.action == 'createClientCert':
            # Create the client cert
            manager = CertificateManager(args.basePath)
            manager.createClientCertificate(args.name, not args.usePass, args.keyLength, args.days)
            return 0
        else:
            logger.error('Unknown argument action [%s]', args.action)
            return 1
    except ValueError as error:
        logger.error(error.message)
        return 1
    except KeyboardInterrupt:
        logger.error('User interrupted')
        return 1
    except:
        logger.exception('Unhandled exception occurred')
        return 1

sys.exit(main())

